@@ -145,7 +145,6 @@ class ChatControl(EventHelper):
         self._scrolled_view.add_command_output(text, is_error)
 
     def add_info_message(self, text: str, timestamp: dt.datetime | None = None) -> None:
-
         self._scrolled_view.add_info_message(text, timestamp)
 
     def drag_data_file_transfer(self, paths: list[str]) -> None:
@@ -199,7 +198,8 @@ class ChatControl(EventHelper):
             messages = app.storage.archive.get_conversation_around_timestamp(
                 self._contact.account, self._contact.jid, timestamp
             )
-            self._add_messages(messages)
+            for message in messages:
+                self._add_message_from_storage(message)
             self._scrolled_view.set_history_complete(False, False)
 
         GLib.idle_add(self._scrolled_view.block_signals, False)
@@ -208,6 +208,14 @@ class ChatControl(EventHelper):
     def mark_as_read(self) -> None:
         self._jump_to_end_button.reset_unread_count()
 
+    def copy_selected_messages(self) -> None:
+        # Message selection mode is active
+        self._scrolled_view.copy_selected_messages()
+        self._message_selection.set_visible(False)
+
+    def message_selection_active(self) -> bool:
+        return self._message_selection.get_visible()
+
     def process_escape(self) -> bool:
         message_selection_active = self._message_selection.get_visible()
         if message_selection_active:
@@ -219,7 +227,6 @@ class ChatControl(EventHelper):
     def switch_contact(
         self, contact: BareContact | GroupchatContact | GroupchatParticipant
     ) -> None:
-
         log.info("Switch to %s (%s)", contact.jid, contact.account)
         if self._contact is not None:
             self._contact.disconnect_all_from_obj(self)
@@ -279,7 +286,11 @@ class ChatControl(EventHelper):
                 return
 
             muc_data = self._client.get_module("MUC").get_muc_data(contact.jid)
-            if muc_data is not None and muc_data.subject is not None:
+            if (
+                muc_data is not None
+                and muc_data.subject is not None
+                and muc_data.subject.text
+            ):
                 self._scrolled_view.add_muc_subject(
                     muc_data.subject, muc_data.last_subject_timestamp
                 )
@@ -397,7 +408,7 @@ class ChatControl(EventHelper):
         if not self._is_event_processable(event):
             return
 
-        self._scrolled_view.set_read_marker(event.marker_id)
+        self._scrolled_view.update_displayed_markers(event)
 
     def _on_reaction_updated(self, event: events.ReactionUpdated) -> None:
         if not self._is_event_processable(event):
@@ -418,7 +429,6 @@ class ChatControl(EventHelper):
         self._scrolled_view.update_call_rows()
 
     def _on_jingle_request_received(self, event: events.JingleRequestReceived) -> None:
-
         if not self._is_event_processable(event):
             return
 
@@ -434,7 +444,6 @@ class ChatControl(EventHelper):
     def _on_file_request_event(
         self, event: events.FileRequestReceivedEvent | events.FileRequestSent
     ) -> None:
-
         if not self._is_event_processable(event):
             return
 
@@ -462,7 +471,6 @@ class ChatControl(EventHelper):
     def _on_autoscroll_changed(
         self, _widget: ConversationView, autoscroll: bool
     ) -> None:
-
         if not autoscroll:
             self._jump_to_end_button.toggle(True)
             return
@@ -479,7 +487,6 @@ class ChatControl(EventHelper):
     def _on_activate_message_selection(
         self, _action: Gio.SimpleAction, param: GLib.Variant
     ) -> None:
-
         pk = param.get_uint32()
         self._scrolled_view.enable_row_selection(pk)
         self._message_selection.set_visible(True)
@@ -489,12 +496,11 @@ class ChatControl(EventHelper):
         self._message_selection.set_visible(False)
 
     def _on_copy_selection(self, _widget: MessageSelection) -> None:
-        self._scrolled_view.copy_selected_messages()
+        self.copy_selected_messages()
 
     def _on_jump_to_message(
         self, _action: Gio.SimpleAction, param: GLib.Variant
     ) -> None:
-
         pk, timestamp = param.unpack()
         self.scroll_to_message(pk, dt.datetime.fromtimestamp(timestamp, dt.UTC))
 
@@ -518,9 +524,8 @@ class ChatControl(EventHelper):
             self._scrolled_view.add_call_message(event=event)
 
     def _add_message(self, message: Message) -> None:
-        # TODO: Unify with _add_db_row()
         if self._allow_add_message():
-            self._scrolled_view.add_message_from_db(message)
+            self._scrolled_view.add_message(message)
 
             if not self.view_is_at_bottom():
                 if message.direction == ChatDirection.OUTGOING:
@@ -530,7 +535,7 @@ class ChatControl(EventHelper):
         else:
             self._jump_to_end_button.add_unread_count()
 
-    def _add_db_row(self, message: Message):
+    def _add_message_from_storage(self, message: Message):
         if message.filetransfers:
             self._scrolled_view.add_jingle_file_transfer(message=message)
             return
@@ -539,11 +544,7 @@ class ChatControl(EventHelper):
             self._scrolled_view.add_call_message(message=message)
             return
 
-        self._scrolled_view.add_message_from_db(message)
-
-    def _add_messages(self, messages: list[Message]):
-        for msg in messages:
-            self._add_db_row(msg)
+        self._scrolled_view.add_message(message)
 
     def _request_messages(self, before: bool) -> Sequence[Message]:
         if before:
@@ -595,7 +596,7 @@ class ChatControl(EventHelper):
         assert self._contact is not None
         for row in rows:
             if not isinstance(row, events.ApplicationEvent):
-                self._add_messages([row])
+                self._add_message_from_storage(row)
 
             elif isinstance(row, events.MUCUserJoined):
                 self._process_muc_user_joined(row)
@@ -653,7 +654,6 @@ class ChatControl(EventHelper):
         event_rows: list[events.ApplicationEvent],
         before: bool,
     ) -> list[HistoryRowT]:
-
         def sort_func(obj: HistoryRowT) -> float:
             return obj.timestamp  # pyright: ignore
 
@@ -671,11 +671,9 @@ class ChatControl(EventHelper):
         _old_contact: types.GroupchatParticipant,
         _new_contact: types.GroupchatParticipant,
     ) -> None:
-
         self._process_muc_nickname_changed(event)
 
     def _process_muc_nickname_changed(self, event: events.MUCNicknameChanged) -> None:
-
         if event.is_self:
             message = _("You are now known as %s") % event.new_name
         else:
@@ -687,7 +685,6 @@ class ChatControl(EventHelper):
     def _on_room_kicked(
         self, _contact: GroupchatContact, _signal_name: str, event: events.MUCRoomKicked
     ) -> None:
-
         self._process_muc_room_kicked(event)
 
     def _process_muc_room_kicked(self, event: events.MUCRoomKicked) -> None:
@@ -746,7 +743,6 @@ class ChatControl(EventHelper):
         user_contact: GroupchatParticipant,
         event: events.MUCUserAffiliationChanged,
     ) -> None:
-
         self._process_muc_user_affiliation_changed(event)
 
     def _process_muc_user_affiliation_changed(
@@ -847,11 +843,9 @@ class ChatControl(EventHelper):
         user_contact: GroupchatParticipant,
         event: events.MUCUserRoleChanged,
     ) -> None:
-
         self._process_muc_user_role_changed(event)
 
     def _process_muc_user_role_changed(self, event: events.MUCUserRoleChanged) -> None:
-
         uf_role = get_uf_role(event.role)
         nick = event.nick
 
@@ -887,11 +881,9 @@ class ChatControl(EventHelper):
         user_contact: GroupchatParticipant,
         event: events.MUCUserHatsChanged,
     ) -> None:
-
         self._process_muc_user_hats_changed(event)
 
     def _process_muc_user_hats_changed(self, event: events.MUCUserHatsChanged) -> None:
-
         if event.hats:
             hats = ", ".join(event.hats)
             if event.is_self:
@@ -916,7 +908,6 @@ class ChatControl(EventHelper):
         _user_contact: GroupchatParticipant,
         event: events.MUCUserStatusShowChanged,
     ) -> None:
-
         self._process_muc_user_status_show_changed(event)
 
     def _on_participant_status_show_changed(
@@ -925,13 +916,11 @@ class ChatControl(EventHelper):
         _signal_name: str,
         event: events.MUCUserStatusShowChanged,
     ) -> None:
-
         self._process_muc_user_status_show_changed(event)
 
     def _process_muc_user_status_show_changed(
         self, event: events.MUCUserStatusShowChanged
     ) -> None:
-
         if isinstance(self._contact, GroupchatContact):
             contact = self._contact
         elif isinstance(self._contact, GroupchatParticipant):
@@ -963,13 +952,11 @@ class ChatControl(EventHelper):
         _signal_name: str,
         event: events.MUCRoomConfigChanged,
     ) -> None:
-
         self._process_muc_room_config_changed(event)
 
     def _process_muc_room_config_changed(
         self, event: events.MUCRoomConfigChanged
     ) -> None:
-
         # http://www.xmpp.org/extensions/xep-0045.html#roomconfig-notify
         status_codes = event.status_codes
         changes: list[str] = []
@@ -1017,7 +1004,6 @@ class ChatControl(EventHelper):
     def _process_muc_room_config_finished(
         self, event: events.MUCRoomConfigFinished
     ) -> None:
-
         self.add_info_message(_("A new group chat has been created"), event.timestamp)
 
     def _on_room_presence_error(
@@ -1026,13 +1012,11 @@ class ChatControl(EventHelper):
         _signal_name: str,
         event: events.MUCRoomPresenceError,
     ) -> None:
-
         self._process_muc_room_presence_error(event)
 
     def _process_muc_room_presence_error(
         self, event: events.MUCRoomPresenceError
     ) -> None:
-
         self.add_info_message(_("Error: %s") % event.error, event.timestamp)
 
     def _on_room_destroyed(
@@ -1041,11 +1025,9 @@ class ChatControl(EventHelper):
         _signal_name: str,
         event: events.MUCRoomDestroyed,
     ) -> None:
-
         self._process_muc_room_destroyed(event)
 
     def _process_muc_room_destroyed(self, event: events.MUCRoomDestroyed) -> None:
-
         reason = event.reason
         reason = "" if reason is None else _("(reason: %s)") % reason
 
@@ -1065,7 +1047,6 @@ class ChatControl(EventHelper):
         _signal_name: str,
         event: events.MUCRoomVoiceRequestError,
     ) -> None:
-
         self._process_muc_room_voice_request_error(event)
 
     def _process_muc_room_voice_request_error(
@@ -1080,7 +1061,6 @@ class ChatControl(EventHelper):
         _user_contact: GroupchatParticipant,
         event: events.MUCUserJoined,
     ) -> None:
-
         self._process_muc_user_joined(event)
 
     def _process_muc_user_joined(self, event: events.MUCUserJoined) -> None:
@@ -1106,8 +1086,7 @@ class ChatControl(EventHelper):
 
         if StatusCode.NICKNAME_MODIFIED in status_codes:
             message = _(
-                "The server has assigned or modified your "
-                "nickname in this group chat"
+                "The server has assigned or modified your nickname in this group chat"
             )
 
         if message is not None:
@@ -1120,7 +1099,6 @@ class ChatControl(EventHelper):
         _user_contact: GroupchatParticipant,
         event: events.MUCUserLeft,
     ) -> None:
-
         self._process_muc_user_left(event)
 
     def _process_muc_user_left(self, event: events.MUCUserLeft) -> None:
@@ -1176,6 +1154,8 @@ class ChatControl(EventHelper):
     def _on_room_subject(
         self, contact: GroupchatContact, _signal_name: str, subject: MucSubject
     ) -> None:
+        if not subject.text:
+            return
 
         if app.settings.get("show_subject_on_join") or not contact.is_joining:
             self._scrolled_view.add_muc_subject(subject)
